// SPDX-License-Identifier: BSD-3-Clause

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include <fcntl.h>
#include <unistd.h>

#include "cmd.h"
#include "utils.h"

#define READ		0
#define WRITE		1

static bool sanity_check_simple_command(simple_command_t *s);
static bool sanity_check_command(command_t *c);
static void clean_resources(int count, ...);
/**
 * Internal change-directory command.
 */
static bool shell_cd(word_t *dir)
{
	char *path = NULL;
	bool dynamically_allocated = false;

	if (dir) {
		path = get_word(dir);
		dynamically_allocated = true;
	} else {
		path = getenv("HOME");
	}

	if (chdir(path) == -1) {
		fprintf(stderr, "cd: %s: %s\n", path, strerror(errno));
		if (dynamically_allocated)
			free(path);
		return false;
	}

	if (dynamically_allocated)
		free(path);
	return true;
}

/**
 * Internal exit/quit command.
 */
static int shell_exit(void)
{
	/* TODO: Execute exit/quit. */
	return SHELL_EXIT;
	// return 0; /* TODO: Replace with actual exit code. */
}

static void handle_input_redirection(word_t *in)
{
	if (!in)
		return;

	char *input_file = get_word(in);
	int fd_in = open(input_file, O_RDONLY);

	DIE(fd_in == -1, "Failed to open input file");

	DIE(dup2(fd_in, STDIN_FILENO) == -1, "Failed to duplicate file descriptor for input redirection");

	clean_resources(1, fd_in, input_file);
}

static void handle_output_redirection(word_t *out, int io_flags, word_t *err)
{
	if (!out && !err)
		return;

	char *output_file = out ? get_word(out) : NULL;
	char *error_file = err ? get_word(err) : NULL;

	//if both stdout and stderr are redirected to the same file
	if (output_file && error_file && strcmp(output_file, error_file) == 0) {
		int flags = O_WRONLY | O_CREAT | ((io_flags & IO_OUT_APPEND) ? O_APPEND : O_TRUNC);
		int fd = open(output_file, flags, 0644);

		DIE(fd == -1, "Failed to open output file for stdout and stderr");

		DIE(dup2(fd, STDOUT_FILENO) == -1, "Failed to redirect stdout");
		DIE(dup2(fd, STDERR_FILENO) == -1, "Failed to redirect stderr");

		clean_resources(1, fd, output_file);
		if (error_file != output_file)
			free(error_file);
		return;
	}

	// if stdout and stderr are redirected to different files
	// initialise with -1 to indicate that file descriptors are not yet opened
	int fd_out = -1, fd_err = -1;

	if (output_file) {
		int flags = O_WRONLY | O_CREAT | ((io_flags & IO_OUT_APPEND) ? O_APPEND : O_TRUNC);

		fd_out = open(output_file, flags, 0644);

		DIE(fd_out == -1, "Failed to open output file");
		DIE(dup2(fd_out, STDOUT_FILENO) == -1, "Failed to redirect stdout");
	}

	if (error_file) {
		int flags = O_WRONLY | O_CREAT | ((io_flags & IO_ERR_APPEND) ? O_APPEND : O_TRUNC);

		fd_err = open(error_file, flags, 0644);

		DIE(fd_err == -1, "Failed to open error file");
		DIE(dup2(fd_err, STDERR_FILENO) == -1, "Failed to redirect stderr");
	}

	clean_resources(2, fd_out, output_file, fd_err, error_file);
}

static void handle_redirections(simple_command_t *s)
{
	if (!s)
		return;

	handle_input_redirection(s->in);
	handle_output_redirection(s->out, s->io_flags, s->err);
}

/**
 * Parse a simple command (internal, environment variable assignment,
 * external command).
 */

void handle_builtin_output(word_t *path)
{
	if (path) {
		char *resolved_path = get_word(path);
		int fd = open(resolved_path, O_WRONLY | O_TRUNC | O_CREAT, 0644);

		if (fd != -1)
			close(fd);
		free(resolved_path);
	}
}


static int parse_simple(simple_command_t *s, int level, command_t *father)
{
	/* TODO: Sanity checks. */
	if (!sanity_check_simple_command(s))
		return -1;

	/* TODO: If builtin command, execute the command. */
	char *command = get_word(s->verb);

	if (strcmp(command, "cd") == 0) {
		clean_resources(1, -1, command);

		handle_builtin_output(s->out);
		handle_builtin_output(s->err);

		if (shell_cd(s->params))
			return 0;
		else
			return -1;
	}

	if (strcmp(command, "exit") == 0 || strcmp(command, "quit") == 0) {
		handle_redirections(s);
		clean_resources(1, -1, command);
		return shell_exit();
	}

	/* TODO: If variable assignment, execute the assignment and return
	 * the exit status.
	 */

	if (strchr(command, '=') != NULL) {
		char *saveptr;
		char *key = strtok_r(command, "=", &saveptr);
		char *value = strtok_r(NULL, "=", &saveptr);

		if (key == NULL || value == NULL) {
			fprintf(stderr, "Invalid environment variable assignment.\n");
			clean_resources(1, -1, command);
			return -1;
		}

		if (setenv(key, value, 1) == -1) {
			fprintf(stderr, "Failed to set environment variable %s.\n", key);
			clean_resources(1, -1, command);
			return -1;
		}
		clean_resources(1, -1, command);
		return 0;
	}


	/* TODO: If external command:
	 *   1. Fork new process
	 *     2c. Perform redirections in child
	 *     3c. Load executable in child
	 *   2. Wait for child
	 *   3. Return exit status
	 */

	//handle external commands
	pid_t pid = fork();

	DIE(pid == -1, "Fork failed");

	if (pid == 0) { //pid for child Process
		handle_redirections(s);

		//prepare arguments for execvp
		int argc = 0;
		char **argv = get_argv(s, &argc);

		execvp(command, argv);
		fprintf(stderr, "Execution failed for '%s'\n", command);

		//free resources and exit child process
		for (int i = 0; i < argc; i++)
			clean_resources(1, -1, argv[i]);
		clean_resources(2, -1, argv, -1, command);
		exit(EXIT_FAILURE);
	}

	//parent process
	clean_resources(1, -1, command);
	int status;

	waitpid(pid, &status, 0);

	if (WIFEXITED(status))
		return WEXITSTATUS(status);
	return -1;

	/* TODO: Replace with actual exit status. */
}

/**
 * Process two commands in parallel, by creating two children.
 */
static bool run_in_parallel(command_t *cmd1, command_t *cmd2, int level, command_t *father)
{
	/* TODO: Execute cmd1 and cmd2 simultaneously. */
	pid_t pids[2]; //pid array for storing child PIDs
	int statuses[2]; //status array for storing child statuses

	//fork cmd1
	pids[0] = fork();
	DIE(pids[0] < 0, "Failed to fork for cmd1");

	if (pids[0] == 0) {
		//child process for cmd1
		int status = parse_command(cmd1, level + 1, father);

		exit(status);
	}

	//fork cmd2
	pids[1] = fork();
	DIE(pids[1] < 0, "Failed to fork for cmd2");

	if (pids[1] == 0) {
		//child process for cmd2
		int status = parse_command(cmd2, level + 1, father);

		exit(status);
	}

	//waiting for both children
	for (int i = 0; i < 2; i++)
		waitpid(pids[i], &statuses[i], 0);

	//return true if both commands succeed
	return WIFEXITED(statuses[0]) && WIFEXITED(statuses[1]) &&
		WEXITSTATUS(statuses[0]) == 0 && WEXITSTATUS(statuses[1]) == 0;

	// return true; /* TODO: Replace with actual exit status. */
}

/**
 *  Helper function to handle child process logic
 **/
static void handle_child_process(command_t *cmd, int pipefd[2], int std_fd, int level, command_t *father)
{
	//redirect the appropriate pipe to the specified standard file descriptor
	if (std_fd == STDOUT_FILENO)
		close(pipefd[READ]);
	else
		close(pipefd[WRITE]);

	DIE(dup2(pipefd[std_fd], std_fd) == -1, "dup2 pipefd failed");
	close(pipefd[std_fd]);

	//parse the command and exit with its status
	int status = parse_command(cmd, level + 1, father);

	exit(status);
}

/**
 * Run commands by creating an anonymous pipe (cmd1 | cmd2).
 */
static bool run_on_pipe(command_t *cmd1, command_t *cmd2, int level, command_t *father)
{
	int pipefd[2];

	DIE(pipe(pipefd) == -1, "Failed to create pipe");

	pid_t pids[2]; //array for storing child PIDs
	int statuses[2]; //array for storing child statuses

	//fork the first child process
	pids[0] = fork();
	DIE(pids[0] == -1, "Failed to fork for cmd1");

	if (pids[0] == 0) {
		//first child process
		handle_child_process(cmd1, pipefd, STDOUT_FILENO, level, father);
	}

	//fork the second child process
	pids[1] = fork();
	DIE(pids[1] == -1, "Failed to fork for cmd2");

	if (pids[1] == 0) {
		// Second child process
		handle_child_process(cmd2, pipefd, STDIN_FILENO, level, father);
	}

	// Close the pipe file descriptors
	close(pipefd[READ]);
	close(pipefd[WRITE]);

	//wait for both child processes to finish
	for (int i = 0; i < 2; i++)
		waitpid(pids[i], &statuses[i], 0);

	//return the exit status of the second command
	if (WIFEXITED(statuses[1]))
		return WEXITSTATUS(statuses[1]);

	return -1;
}



/**
 * Parse and execute a command.
 */
int parse_command(command_t *c, int level, command_t *father)
{
	/* TODO: sanity checks */
	if (!sanity_check_command(c))
		return -1;

	if (c->op == OP_NONE) {
		/* TODO: Execute a simple command. */
		return parse_simple(c->scmd, level, father);
		// return 0; /* TODO: Replace with actual exit code of command. */
	}

	switch (c->op) {
	case OP_SEQUENTIAL:
		parse_command(c->cmd1, level, c);
		return parse_command(c->cmd2, level, c);
		/* TODO: Execute the commands one after the other. */
	case OP_PARALLEL:
		/* TODO: Execute the commands simultaneously. */
		return run_in_parallel(c->cmd1, c->cmd2, level, c);
		// break;

	case OP_CONDITIONAL_NZERO:
		/* TODO: Execute the second command only if the first one
		 * returns non zero.
		 */
		if (parse_command(c->cmd1, level, c) != 0)
			return parse_command(c->cmd2, level, c);
		return 0;
	case OP_CONDITIONAL_ZERO:
		/* TODO: Execute the second command only if the first one
		 * returns zero.
		 */
		if (parse_command(c->cmd1, level, c) == 0)
			return parse_command(c->cmd2, level, c);
		return 0;
	case OP_PIPE:
		/* TODO: Redirect the output of the first command to the
		 * input of the second.
		 */
		return run_on_pipe(c->cmd1, c->cmd2, level, c);

	default:
		return SHELL_EXIT;
	}

	return 0; /* TODO: Replace with actual exit code of command. */
}

static void clean_resources(int count, ...)
{
	va_list ap;

	va_start(ap, count);

	for (int i = 0; i < count; i++) {
		int  fd     = va_arg(ap, int);
		char *fname = va_arg(ap, char *);

		if (fd >= 0)
			close(fd);

		if (fname)
			free(fname);
	}

	va_end(ap);
}

static bool sanity_check_simple_command(simple_command_t *s)
{
	if (s == NULL)
		return false;

	if (s->verb == NULL)
		return false;

	return true;
}

static bool sanity_check_command(command_t *c)
{
	if (c == NULL)
		return false;

	if (c->op == OP_NONE && c->scmd == NULL)
		return false;

	if ((c->op != OP_NONE) && (c->cmd1 == NULL || c->cmd2 == NULL))
		return false;

	return true;
}
